// Qt headers
#include <QApplication>
#include <QLineEdit>

// gemc headers
#include "gemc_MainGui.h"
#include "string_utilities.h"

// All icon images must be 96x84
#include "images/run_control_xpm.h"
#include "images/camera_control_xpm.h"
#include "images/detector_inspector_xpm.h"
#include "images/infos_xpm.h"
#include "images/g4dialog_xpm.h"
#include "images/analog_signal_xpm.h"
#include "images/physics_xpm.h"

// C++ headers
#include <iostream>
#include <string>
using namespace std;

gemcMainWidget::gemcMainWidget(goptions *Opts, G4RunManager *rm,  map<string, sensitiveDetector*> SDM, map<string, detector> *hMap, map<string, G4Material*> matMap)
{
	gemcOpt    = Opts;
	RM = rm;

	// revisit this, why we need it?
	double qt_mode = gemcOpt->optMap["USE_GUI"].arg ;
	
	contentsWidget = new QListWidget;
	contentsWidget->setViewMode(QListView::IconMode);
	
	// icon size
	contentsWidget->setIconSize(QSize(80, 80));
	contentsWidget->setMovement(QListView::Static);

	// icon container sizes
	contentsWidget->setMinimumWidth(110);
	contentsWidget->setMaximumWidth(110);
	contentsWidget->setMinimumHeight(710);
	contentsWidget->setMaximumHeight(710);
	
	// makes all icon the same size
	contentsWidget->setUniformItemSizes(1);
	
	// activate tracking (for mouse over effect)
	contentsWidget->setMouseTracking(1);
	
	contentsWidget->setSpacing(8);

	// Content:
	// Run Control
	// Camera Control
	// Infos
	// G4 Dialog
	// Oscilloscope Signal
	// Physics List Control
  
	pagesWidget = new QStackedWidget;
	pagesWidget->addWidget(new run_control           (this, gemcOpt));
	pagesWidget->addWidget(new camera_control        (this, gemcOpt));
	pagesWidget->addWidget(new detector_tree         (this, *gemcOpt, RM, hMap,  matMap));
	pagesWidget->addWidget(new infos                 (this, gemcOpt));
	pagesWidget->addWidget(new g4dialog              (this, gemcOpt));
	pagesWidget->addWidget(gsig = new gsignal        (this, gemcOpt, SDM));
	pagesWidget->setMinimumWidth(650);
	pagesWidget->setMaximumWidth(650);
	pagesWidget->setMinimumHeight(750);
	pagesWidget->setMaximumHeight(750);
	QPushButton *closeButton = new QPushButton(tr("Exit"));
	
	createIcons();
	contentsWidget->setCurrentRow(0);
	
	// revisit this, why we need it?
	if(qt_mode > 1)
		contentsWidget->setCurrentRow(1);
	
	connect(closeButton, SIGNAL(clicked()), this, SLOT(gemc_quit()));
	
	QHBoxLayout *horizontalLayout = new QHBoxLayout;
	horizontalLayout->addWidget(contentsWidget);
	horizontalLayout->addWidget(pagesWidget, 1);
	
	QHBoxLayout *buttonsLayout = new QHBoxLayout;
	buttonsLayout->addStretch(1);
	buttonsLayout->addWidget(closeButton);
	
	QVBoxLayout *mainLayout = new QVBoxLayout;
	mainLayout->addLayout(horizontalLayout);
	mainLayout->addLayout(buttonsLayout);
	setLayout(mainLayout);
	
	setWindowTitle(tr("Config Dialog"));

	return;
}



void gemcMainWidget::changePage(QListWidgetItem *current, QListWidgetItem *previous)
{
	if (!current)
		current = previous;
	
	// manually setting the content index
	if(contentsWidget->row(current) == 0) pagesWidget->setCurrentIndex(0);
	if(contentsWidget->row(current) == 1) pagesWidget->setCurrentIndex(1);
	if(contentsWidget->row(current) == 2) pagesWidget->setCurrentIndex(2);
	if(contentsWidget->row(current) == 3) pagesWidget->setCurrentIndex(3);
	if(contentsWidget->row(current) == 4) pagesWidget->setCurrentIndex(4);
	if(contentsWidget->row(current) == 5)
	{
		pagesWidget->setCurrentIndex(5);
		gsig->CreateSDetsTree();
	}
	if(contentsWidget->row(current) == 6) pagesWidget->setCurrentIndex(6);
}

void gemcMainWidget::createIcons()
{
	i_default = QColor( 160,  180, 200);
	i_hovered = QColor( 124,  164, 210);
	
	
	
	buttons.push_back(addItem("Run Control", QPixmap(run_control_xpm)));
	buttons.push_back(addItem("Camera",      QPixmap(camera_control_xpm)));
	buttons.push_back(addItem("Detector",    QPixmap(detector_inspector_xpm)));
	buttons.push_back(addItem("Infos",       QPixmap(infos_xpm)));
	buttons.push_back(addItem("G4Dialog",    QPixmap(g4dialog_xpm)));
	buttons.push_back(addItem("Signals",     QPixmap(analog_signal_xpm)));
	buttons.push_back(addItem("Physics",     QPixmap(physics_xpm)));

		
	connect(contentsWidget,
					SIGNAL(currentItemChanged(QListWidgetItem *, QListWidgetItem*)),
          this, SLOT(changePage(QListWidgetItem *, QListWidgetItem*)));
	
	connect(contentsWidget,	SIGNAL(itemEntered(QListWidgetItem *)), this, SLOT(change_background(QListWidgetItem *)) );
	
}

void gemcMainWidget::change_background(QListWidgetItem* item)
{
	for(int i=0; i<contentsWidget->count(); i++)
		contentsWidget->item(i)->setBackground(QBrush(i_default));
	
	item->setBackground(QBrush(i_hovered));

}



QListWidgetItem* gemcMainWidget::addItem(string text, QPixmap pixmap)
{
	QListWidgetItem *newItem = new QListWidgetItem(contentsWidget);
	newItem->setIcon(pixmap);
	newItem->setText(tr(text.c_str()));
	newItem->setBackground(QBrush(i_default));
	newItem->setTextAlignment(Qt::AlignHCenter);
	newItem->setFlags( Qt::ItemIsSelectable | Qt::ItemIsEnabled);
	
	return newItem;
}


gemcMainWidget::~gemcMainWidget()
{
	string hd_msg = gemcOpt->optMap["LOG_MSG"].args;


	cout << endl;
}





